@@ -1,4 +1,19 @@
-/*
+/*  Copyright 1995-1998,2000-2003,2005,2007-2009 Alain Knaff.
+ *  This file is part of mtools.
+ *
+ *  Mtools is free software: you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
+ *
+ *  Mtools is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with Mtools.  If not, see <http://www.gnu.org/licenses/>.
+ *
  * mk_direntry.c
  * Make new directory entries, and handles name clashes
  *
@@ -16,11 +31,43 @@
 #include "fs.h"
 #include "stream.h"
 #include "mainloop.h"
+#include "file_name.h"
 
-static __inline__ int ask_rename(ClashHandling_t *ch,
-			     char *longname, int isprimary, char *argname)
+/**
+ * Converts input to shortname
+ * @param un unix name (in Unix charset)
+ * 
+ * @return 1 if name had to be mangled
+ */
+static __inline__ int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
+					   const char *un, dos_name_t *dn)
+{
+	int mangled;
+
+	/* Then do conversion to dn */
+	ch->name_converter(cp, un, 0, &mangled, dn);
+	dn->sentinel = '\0';
+	if (dn->base[0] == '\xE5')
+		dn->base[0] = '\x05';
+	return mangled;
+}
+
+static __inline__ void chomp(char *line)
+{
+	int l = strlen(line);
+	while(l > 0 && (line[l-1] == '\n' || line[l-1] == '\r')) {
+		line[--l] = '\0';
+	}
+}
+
+/**
+ * Asks for an alternative new name for a file, in case of a clash
+ */
+static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
+				 dos_name_t *shortname,
+				 char *longname,
+				 int isprimary)
 {
-	char shortname[13];
 	int mangled;
 
 	/* TODO: Would be nice to suggest "autorenamed" version of name, press 
@@ -38,32 +85,64 @@ static __inline__ int ask_rename(ClashHandling_t *ch,
 
 	mangled = 0;
 	do {
+		char tname[4*MAX_VNAMELEN+1];
 		fprintf(stderr, "New %s name for \"%s\": ",
 			isprimary ? "primary" : "secondary", longname);
 		fflush(stderr);
-		if (! fgets(name, maxsize, opentty(0)))
+		if (! fgets(tname, 4*MAX_VNAMELEN+1, opentty(0)))
 			return 0;
-
-		/* Eliminate newline(s) in the file name */
-		name[strlen(name)-1]='\0';
-		if (!isprimary)
-			ch->name_converter(shortname,0, &mangled, argname);
+		chomp(tname);
+		if (isprimary)
+			strcpy(longname, tname);
+		else
+			mangled = convert_to_shortname(cp, 
+						       ch, tname, shortname);
 	} while (mangled & 1);
 	return 1;
 #undef maxsize
 #undef name
 }
 
-static __inline__ clash_action ask_namematch(char *name, int isprimary, 
-					     ClashHandling_t *ch, 
+/**
+ * This function determines the action to be taken in case there is a problem
+ * with target name (clash, illegal characters, or reserved)
+ * The decision either comes from the default (ch), or the user will be
+ * prompted if there is no default
+ */
+static __inline__ clash_action ask_namematch(doscp_t *cp,
+					     dos_name_t *dosname,
+					     char *longname,
+					     int isprimary, 
+					     ClashHandling_t *ch,
 					     int no_overwrite,
 					     int reason)
 {
+	/* User's answer letter (from keyboard). Only first letter is used,
+	 * but we allocate space for 10 in order to account for extra garbage
+	 * that user may enter
+	 */
 	char ans[10];
-	clash_action a;
-	int perm;
-	char unix_shortname[13];
 
+	/**
+	 * Return value: action to be taken
+	 */
+	clash_action a;
+
+	/**
+	 * Should this decision be made permanent (do no longer ask same
+	 * question)
+	 */
+	int perm;
+
+	/**
+	 * Buffer for shortname
+	 */
+	char name_buffer[4*13];
+
+	/**
+	 * Name to be printed
+	 */
+	char *name;
 
 #define EXISTS 0
 #define RESERVED 1
@@ -74,10 +153,6 @@ static __inline__ clash_action ask_namematch(char *name, int isprimary,
 		"is reserved",
 		"contains illegal character(s)"};
 
-
-	if (!isprimary)
-		name = unix_normalize(unix_shortname, name, name+8);
-
 	a = ch->action[isprimary];
 
 	if(a == NAMEMATCH_NONE && !opentty(1)) {
@@ -85,6 +160,11 @@ static __inline__ clash_action ask_namematch(char *name, int isprimary,
 		return NAMEMATCH_SKIP;
 	}
 
+	if (!isprimary)
+		name = unix_normalize(cp, name_buffer, dosname);
+	else
+		name = longname;
+
 	perm = 0;
 	while (a == NAMEMATCH_NONE) {
 		fprintf(stderr, "%s file name \"%s\" %s.\n",
@@ -109,7 +189,8 @@ static __inline__ clash_action ask_namematch(char *name, int isprimary,
 			else
 				ans[0] = rep;
 		} else {
-			fgets(ans, 9, opentty(0));
+			if(fgets(ans, 9, opentty(0)) == NULL)
+				ans[0] = 'q';
 		}
 		perm = isupper((unsigned char)ans[0]);
 		switch(tolower((unsigned char)ans[0])) {
@@ -154,7 +235,12 @@ static __inline__ clash_action ask_namematch(char *name, int isprimary,
 	return a;
 }
 
-/* Returns:
+/*
+ * Processes a name match
+ *  dosname short dosname (ignored if is_primary)
+ *
+ *
+ * Returns:
  * 2 if file is to be overwritten
  * 1 if file was renamed
  * 0 if it was skipped
@@ -165,7 +251,8 @@ static __inline__ clash_action ask_namematch(char *name, int isprimary,
  *
  * Also, immediately copy the original name so that messages can use it.
  */
-static __inline__ clash_action process_namematch(char *name,
+static __inline__ clash_action process_namematch(doscp_t *cp,
+						 dos_name_t *dosname,
 						 char *longname,
 						 int isprimary,
 						 ClashHandling_t *ch,
@@ -180,7 +267,8 @@ static __inline__ clash_action process_namematch(char *name,
 		name, default_action, ch->ask);
 #endif
 
-	action = ask_namematch(name, isprimary, ch, no_overwrite, reason);
+	action = ask_namematch(cp, dosname, longname,
+			       isprimary, ch, no_overwrite, reason);
 
 	switch(action){
 	case NAMEMATCH_QUIT:
@@ -195,7 +283,7 @@ static __inline__ clash_action process_namematch(char *name,
 		 * new name collision, and b) finding a big enough VSE.
 		 * Change the name, so that it won't collide again.
 		 */
-		ask_rename(ch, longname, isprimary, name);
+		ask_rename(cp, ch, dosname, longname, isprimary);
 		return action;
 	case NAMEMATCH_AUTORENAME:
 		/* Very similar to NAMEMATCH_RENAME, except that we need to
@@ -204,10 +292,10 @@ static __inline__ clash_action process_namematch(char *name,
 		 * keep trying the same one.
 		 */
 		if (isprimary) {
-			autorename_long(name, 1);
+			autorename_long(longname, 1);
 			return NAMEMATCH_PRENAME;
 		} else {
-			autorename_short(name, 1);
+			autorename_short(dosname, 1);
 			return NAMEMATCH_RENAME;
 		}
 	case NAMEMATCH_OVERWRITE:
@@ -220,22 +308,10 @@ static __inline__ clash_action process_namematch(char *name,
 	}
 }
 
-
-static void clear_scan(char *longname, int use_longname, struct scan_state *s)
+static int contains_illegals(const char *string, const char *illegals,
+			     int len)
 {
-	s->shortmatch = s->longmatch = s->slot = -1;
-	s->free_end = s->got_slots = s->free_start = 0;
-
-	if (use_longname & 1)
-                s->size_needed = 2 + (strlen(longname)/VSE_NAMELEN);
-	else
-                s->size_needed = 1;
-}
-
-
-static int contains_illegals(const char *string, const char *illegals)
-{
-	for(; *string ; string++)
+	for(; *string && len--; string++)
 		if((*string < ' ' && *string != '\005' && !(*string & 0x80)) ||
 		   strchr(illegals, *string))
 			return 1;
@@ -265,18 +341,20 @@ static int is_reserved(char *ans, int islong)
 }
 
 static __inline__ clash_action get_slots(Stream_t *Dir,
-					 char *dosname, char *longname,
+					 dos_name_t *dosname,
+					 char *longname,
 					 struct scan_state *ssp,
 					 ClashHandling_t *ch)
 {
 	int error;
 	clash_action ret;
-	int match=0;
+	int match_pos=0;
 	direntry_t entry;
 	int isprimary;
 	int no_overwrite;
 	int reason;
 	int pessimisticShortRename;
+	doscp_t *cp = GET_DOSCONVERT(Dir);
 
 	pessimisticShortRename = (ch->action[0] == NAMEMATCH_AUTORENAME);
 
@@ -286,32 +364,32 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 	   longname[strspn(longname,". ")] == '\0'){
 		reason = RESERVED;
 		isprimary = 1;
-	} else if(contains_illegals(longname,long_illegals)) {
+	} else if(contains_illegals(longname,long_illegals,1024)) {
 		reason = ILLEGALS;
 		isprimary = 1;
-	} else if(is_reserved(dosname,0)) {
+	} else if(is_reserved(dosname->base,0)) {
 		reason = RESERVED;
 		ch->use_longname = 1;
 		isprimary = 0;
-	} else if(contains_illegals(dosname,short_illegals)) {
+	} else if(contains_illegals(dosname->base,short_illegals,11)) {
 		reason = ILLEGALS;
 		ch->use_longname = 1;
 		isprimary = 0;
 	} else {
 		reason = EXISTS;
-		clear_scan(longname, ch->use_longname, ssp);
 		switch (lookupForInsert(Dir,
 					&entry,
 					dosname, longname, ssp,
-					ch->ignore_entry, 
+					ch->ignore_entry,
 					ch->source_entry,
-					pessimisticShortRename && 
+					pessimisticShortRename &&
+					ch->use_longname,
 					ch->use_longname)) {
 			case -1:
 				return NAMEMATCH_ERROR;
 				
 			case 0:
-				return NAMEMATCH_SKIP; 
+				return NAMEMATCH_SKIP;
 				/* Single-file error error or skip request */
 				
 			case 5:
@@ -320,54 +398,54 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 				
 			case 6:
 				return NAMEMATCH_SUCCESS; /* Success */
-		}	    
-		match = -2;
+		}	
+		match_pos = -2;
 		if (ssp->longmatch > -1) {
 			/* Primary Long Name Match */
 #ifdef debug
 			fprintf(stderr,
-				"Got longmatch=%d for name %s.\n", 
+				"Got longmatch=%d for name %s.\n",
 				longmatch, longname);
 #endif			
-			match = ssp->longmatch;
+			match_pos = ssp->longmatch;
 			isprimary = 1;
 		} else if ((ch->use_longname & 1) && (ssp->shortmatch != -1)) {
 			/* Secondary Short Name Match */
 #ifdef debug
 			fprintf(stderr,
-				"Got secondary short name match for name %s.\n", 
+				"Got secondary short name match for name %s.\n",
 				longname);
 #endif
 
-			match = ssp->shortmatch;
+			match_pos = ssp->shortmatch;
 			isprimary = 0;
 		} else if (ssp->shortmatch >= 0) {
 			/* Primary Short Name Match */
 #ifdef debug
 			fprintf(stderr,
-				"Got primary short name match for name %s.\n", 
+				"Got primary short name match for name %s.\n",
 				longname);
 #endif
-			match = ssp->shortmatch;
+			match_pos = ssp->shortmatch;
 			isprimary = 1;
-		} else 
+		} else
 			return NAMEMATCH_RENAME;
 
-		if(match > -1) {
-			entry.entry = match;
+		if(match_pos > -1) {
+			entry.entry = match_pos;
 			dir_read(&entry, &error);
 			if (error)
 			    return NAMEMATCH_ERROR;
 			/* if we can't overwrite, don't propose it */
-			no_overwrite = (match == ch->source || IS_DIR(&entry));
+			no_overwrite = (match_pos == ch->source || IS_DIR(&entry));
 		}
 	}
-	ret = process_namematch(isprimary ? longname : dosname, longname,
+	ret = process_namematch(cp, dosname, longname,
 				isprimary, ch, no_overwrite, reason);
 	
-	if (ret == NAMEMATCH_OVERWRITE && match > -1){
+	if (ret == NAMEMATCH_OVERWRITE && match_pos > -1){
 		if((entry.dir.attr & 0x5) &&
-		   (ask_confirmation("file is read only, overwrite anyway (y/n) ? ",0,0)))
+		   (ask_confirmation("file is read only, overwrite anyway (y/n) ? ")))
 			return NAMEMATCH_RENAME;
 		/* Free up the file to be overwritten */
 		if(fatFreeWithDirentry(&entry))
@@ -375,11 +453,11 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 		
 #if 0
 		if(isprimary &&
-		   match - ssp->match_free + 1 >= ssp->size_needed){
+		   match_pos - ssp->match_free + 1 >= ssp->size_needed){
 			/* reuse old entry and old short name for overwrite */
-			ssp->free_start = match - ssp->size_needed + 1;
+			ssp->free_start = match_pos - ssp->size_needed + 1;
 			ssp->free_size = ssp->size_needed;
-			ssp->slot = match;
+			ssp->slot = match_pos;
 			ssp->got_slots = 1;
 			strncpy(dosname, dir.name, 3);
 			strncpy(dosname + 8, dir.ext, 3);
@@ -397,7 +475,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 
 
 static __inline__ int write_slots(Stream_t *Dir,
-				  char *dosname, 
+				  dos_name_t *dosname,
 				  char *longname,
 				  struct scan_state *ssp,
 				  write_data_callback *cb,
@@ -412,8 +490,8 @@ static __inline__ int write_slots(Stream_t *Dir,
 
 	entry.Dir = Dir;
 	entry.entry = ssp->slot;
-	strncpy(entry.name, longname, sizeof(entry.name)-1);
-	entry.name[sizeof(entry.name)-1]='\0';
+	native_to_wchar(longname, entry.name, MAX_VNAMELEN, 0, 0);
+	entry.name[MAX_VNAMELEN]='\0';
 	entry.dir.Case = Case & (EXTCASE | BASECASE);
 	if (cb(dosname, longname, arg, &entry) >= 0) {
 		if ((ssp->size_needed > 1) &&
@@ -425,7 +503,7 @@ static __inline__ int write_slots(Stream_t *Dir,
 			write_vfat(Dir, dosname, 0,
 				   ssp->free_start, &entry);
 		}
-		/* clear_vses(Dir, ssp->free_start + ssp->size_needed, 
+		/* clear_vses(Dir, ssp->free_start + ssp->size_needed,
 		   ssp->free_end); */
 	} else
 		return 0;
@@ -446,19 +524,20 @@ static void stripspaces(char *name)
 }
 
 
-int _mwrite_one(Stream_t *Dir,
-		char *argname,
-		char *shortname,
-		write_data_callback *cb,
-		void *arg,
-		ClashHandling_t *ch)
+static int _mwrite_one(Stream_t *Dir,
+		       char *argname,
+		       char *shortname,
+		       write_data_callback *cb,
+		       void *arg,
+		       ClashHandling_t *ch)
 {
 	char longname[VBUFSIZE];
 	const char *dstname;
-	char dosname[13];
+	dos_name_t dosname;
 	int expanded;
 	struct scan_state scan;
 	clash_action ret;
+	doscp_t *cp = GET_DOSCONVERT(Dir);
 
 	expanded = 0;
 
@@ -475,14 +554,14 @@ int _mwrite_one(Stream_t *Dir,
 	}
 
 	if(shortname){
-		ch->name_converter(shortname,0, &ch->use_longname, dosname);
+		convert_to_shortname(cp, ch, shortname, &dosname);
 		if(ch->use_longname & 1){
 			/* short name mangled, treat it as a long name */
 			argname = shortname;
 			shortname = 0;
 		}
 	}
-						
+
 	if (argname[0] && (argname[1] == ':')) {
 		/* Skip drive letter */
 		dstname = argname + 2;
@@ -494,29 +573,33 @@ int _mwrite_one(Stream_t *Dir,
 	strncpy(longname, dstname, VBUFSIZE-1);
 
 	if(shortname) {
-		ch->name_converter(shortname,0, &ch->use_longname, dosname);
+		ch->use_longname =
+			convert_to_shortname(cp, ch, shortname, &dosname);
 		if(strcmp(shortname, longname))
 			ch->use_longname |= 1;
-	} else
-		ch->name_converter(longname,0, &ch->use_longname, dosname);
+	} else {
+		ch->use_longname =
+			convert_to_shortname(cp, ch, longname, &dosname);
+	}
 
 	ch->action[0] = ch->namematch_default[0];
 	ch->action[1] = ch->namematch_default[1];
 
 	while (1) {
-		switch((ret=get_slots(Dir, dosname, longname,
-				      &scan, ch))){
+		switch((ret=get_slots(Dir, &dosname, longname, &scan, ch))){
 			case NAMEMATCH_ERROR:
-				return -1;	/* Non-file-specific error, 
+				return -1;	/* Non-file-specific error,
 						 * quit */
 				
 			case NAMEMATCH_SKIP:
-				return -1;	/* Skip file (user request or 
+				return -1;	/* Skip file (user request or
 						 * error) */
 
 			case NAMEMATCH_PRENAME:
-				ch->name_converter(longname,0,
-						   &ch->use_longname, dosname);
+				ch->use_longname =
+					convert_to_shortname(cp, ch,
+							     longname,
+							     &dosname);
 				continue;
 			case NAMEMATCH_RENAME:
 				continue;	/* Renamed file, loop again */
@@ -525,9 +608,9 @@ int _mwrite_one(Stream_t *Dir,
 				/* No collision, and not enough slots.
 				 * Try to grow the directory
 				 */
-				if (expanded) {	/* Already tried this 
+				if (expanded) {	/* Already tried this
 						 * once, no good */
-					fprintf(stderr, 
+					fprintf(stderr,
 						"%s: No directory slots\n",
 						progname);
 					return -1;
@@ -539,7 +622,7 @@ int _mwrite_one(Stream_t *Dir,
 				continue;
 			case NAMEMATCH_OVERWRITE:
 			case NAMEMATCH_SUCCESS:
-				return write_slots(Dir, dosname, longname,
+				return write_slots(Dir, &dosname, longname,
 						   &scan, cb, arg,
 						   ch->use_longname);
 			default:
@@ -621,3 +704,8 @@ int handle_clash_options(ClashHandling_t *ch, char c)
 			return -1;
 	}
 }
+
+void dosnameToDirentry(const struct dos_name_t *dn, struct directory *dir) {
+	strncpy(dir->name, dn->base, 8);
+	strncpy(dir->ext, dn->ext, 3);
+}
